#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "sh.h"
#include "shio.h"
#include "shvar.h"
#include "util.h"

/*
 * this could have been an instance of the dynamically sized array
 * with an iterator function.
 */

extern char	**environ;

struct var {
	char	*name, *value;
	char	*env_entry;
} *var_tbl = 0;

int	var_tbl_used = 0, var_tbl_sz = 0;
/* invariant: var_tbl_used <= var_tbl_sz */

void		dollar_update_env(struct var	*vp)
{
	char	*old = vp->env_entry;

	vp->env_entry = safe_malloc(strlen(vp->name)+1+strlen(vp->value)+1);
	sprintf(vp->env_entry,"%s=%s",vp->name,vp->value);
	if (putenv(vp->env_entry)) {
		perror(SHNAME ": updating environment");
		exit(1);
	}
	if (old) free(old);
}

/* ownership of malloc'd name/value pair is passed to var_tbl */
void		dollar_set_var(char	*name,
			       char	*value)
{
	int		i;
	struct var	*ntbl;
	int	     	newsize;

	/* has old value? */
	for (i = 0; i < var_tbl_used; i++) {
		if (!strcmp(var_tbl[i].name,name)) {
			/* overwrite */
			free(var_tbl[i].value);
			var_tbl[i].value = value;
			free(name);	/* BUGFIX */
			if (var_tbl[i].env_entry)
				dollar_update_env(&var_tbl[i]);
			return;
		}
	}
	
	if (var_tbl_used == var_tbl_sz) {
		newsize = 2 * var_tbl_sz;
		ntbl = realloc(var_tbl,newsize * sizeof *ntbl);
		if (!ntbl) {
			perror(SHNAME ": dollar_set_var");
			desc_printf(2,SHNAME ": cannot set %s to %s\n",name,value);
			return;
		}
		var_tbl = ntbl;
	}
	/* var_tbl_used < var_tbl_sz */
	var_tbl[var_tbl_used].name = name;
	var_tbl[var_tbl_used].value = value;
	var_tbl[var_tbl_used].env_entry = 0;
	var_tbl_used++;	/* BUGFIX */
}

int		dollar_mark_exported(char	*name)
{
	int	i;

	for (i = 0; i < var_tbl_used; i++) {
		if (!strcmp(var_tbl[i].name,name)) {
			dollar_update_env(&var_tbl[i]);
			return 1;
		}
	}
	return 0;
}

/* NEW FN */
void		dollar_on_fork(void)
{
	char	pidbuf[16];

	(void) sprintf(pidbuf,"%ld",(unsigned long) getpid());
	dollar_set_var(dupstr("$"),dupstr(pidbuf));
}

void		dollar_init(void)
{
	char	**ep = environ, *var, *eq;

	/* copy the environment to the shell variable list */
	var_tbl = malloc((var_tbl_sz = 64) * sizeof *var_tbl);
	if (!var_tbl) {
		perror(SHNAME ": dollar_init");
		exit(1);
	}
	while ((var = *ep) != 0) {
		eq = strchr(var,'=');
		if (!eq) {
			desc_printf(2,SHNAME ": badly formatted environment entry \"%s\"\n",*ep);
			exit(1);
		}
		/* XXX -- hack */
		*eq = '\0';
		dollar_set_var(dupstr(var),dupstr(eq+1));
		(void) dollar_mark_exported(var);
		*eq = '=';
		ep++;
	}
	dollar_set_var(dupstr("?"),dupstr("0"));
	dollar_on_fork();
}

/*
 * dollar module still owns value; user can copy out.
 */
const char	*dollar_value(char	*name)
{
	int	i;

	for (i = 0; i < var_tbl_used; i++) {
		if (!strcmp(var_tbl[i].name,name)) {
			return var_tbl[i].value;
		}
	}
	return 0;
}

int		dollar_unset(char	*name)
{
	int	i;
	char	*eq;

	for (i = 0; i < var_tbl_used; i++) {
		if (!strcmp(var_tbl[i].name,name)) {
			free(var_tbl[i].value);
			free(var_tbl[i].name);
			/*
			 * XXX: we leak memory if this is an
			 * environment variable: env_data is not
			 * freed; instead, we unset it the best we can
			 * w/o an unsetenv(3) library function (it
			 * exists on BSD systems only).
			 */
			if (var_tbl[i].env_entry) {
				eq = strchr(var_tbl[i].env_entry,'=');
				if (!eq) {
					desc_printf(2,SHNAME": internal error, no equal sign in environment variable string \"%s\".\n",var_tbl[i].env_entry);
					abort();
				}
				eq[1] = 0;
			}
			var_tbl[i] = var_tbl[--var_tbl_used];
			return 1;
		}
	}
	return 0;
}

/*
 * ownership of the newly constructed environment template is passed
 * to caller, who is responsible for freeing it if needed.
 */
char		**dollar_construct_envp(void)
{
	char	**newenv;
	int	env_size;
	int	i, j;

	for (env_size = i = 0; i < var_tbl_used; i++) {
		if (!var_tbl[i].env_entry) continue;
		++env_size;
	}
	newenv = malloc((env_size + 1) * sizeof(*newenv));
	if (!newenv) {
		perror(SHNAME ": dollar_construct_envp");
		desc_printf(2,SHNAME ": no memory for constructing environment table\n");
		exit(1);
	}
	for (j = i = 0; i < var_tbl_used; i++) {
		if (!var_tbl[i].env_entry) continue;
		newenv[j++] = var_tbl[i].env_entry;
	}
	newenv[j] = 0;
	return newenv;
}

void		dollar_free_envp(char	**envp)
{
	free(envp);
}
