#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "sh.h"
#include "util.h"
#include "shio.h"

#define	STRBUF_INITIAL_SIZE	128
#define	DYNBUF_INITIAL_SIZE	8

/*
 * duplicates a string into a new memory arena allocated from the heap.
 * this routine will succeed or die trying.
 */
char		*dupstr(char	*str)
{
	char	*tmp;
	int	len;

	if (!str) return str; /* should not be necessary */
	tmp = malloc((len = strlen(str))+1);
	if (!tmp) {
		perror(SHNAME ": dupstr");
		desc_printf(2,SHNAME ": out of memory duplicating a %d byte string\n",len);
		exit(1);
	}
	strcpy(tmp,str);
	return tmp;
}

/*
 * useful, but doesn't give good error reporting as to where the problem
 * occurred.
 */
void		*safe_malloc(size_t	nbytes)
{
	void	*p = malloc(nbytes);

	if (!p) {
		perror(SHNAME ": safe_malloc");
		exit(1);
	}
	return p;
}

void		strbuf_constr(struct strbuf	*sbp)
{
	sbp->used = 0;
	sbp->size = 0;
	sbp->buf = 0;
}

struct strbuf	*strbuf_constr_ptr(void)
{
	struct strbuf	*sbp = safe_malloc(sizeof *sbp);
	strbuf_constr(sbp);
	return sbp;
}

void		strbuf_destr(struct strbuf	*sbp)
{
	if (sbp->buf) free(sbp->buf);
	sbp->size = sbp->used = 0;
	sbp->buf = 0;
}

void		strbuf_destr_ptr(struct strbuf	*sbp)
{
	strbuf_destr(sbp);
	free(sbp);
}

char		*strbuf_take_buf(struct strbuf	*sbp)
{
	char	*buf;

	buf = sbp->buf;
	sbp->buf = 0;
	sbp->size = sbp->used = 0;
	return buf;
}

void		strbuf_add_ch(struct strbuf	*sbp,
			      char		ch)
{
	char	*nbuf;
	int	nsize;

	if (sbp->used == sbp->size) {
		if (sbp->size == 0)	nsize = STRBUF_INITIAL_SIZE;
		else			nsize = 2 * sbp->size;
		nbuf = realloc(sbp->buf,nsize);
		if (!nbuf) {
			perror("strbuf_add");
			desc_printf(2,SHNAME ": out of memory expanding string buffer\n");
			exit(1);
		}
		/* nbuf != NULL */
		sbp->buf = nbuf;
		sbp->size = nsize;
	}
	/* used < size */
	sbp->buf[sbp->used++] = ch;
}

int		strbuf_nchars(struct strbuf	*sbp)
{
	return sbp->used;
}

void		strbuf_clear(struct strbuf	*sbp)
{
	sbp->used = 0;
}

/*
 * dynbuf is very similar to strbuf, except we are storing
 * void* instead of char.
 *
 * it's basically a dynamically sized (non-sparse) array.
 *
 * in C++ we would have used templates to do both.
 */
void		dynbuf_constr(struct dynbuf	*dbp)
{
	dbp->size = dbp->used = 0;
	dbp->obj = 0;
}

void		dynbuf_destr(struct dynbuf	*dbp)
{
	int	i;

	if (dbp->size) {
		for (i = 0; i < dbp->used; i++) {
			free(dbp->obj[i]);
		}
		free(dbp->obj);
	}
	dbp->obj = 0;
	dbp->size = dbp->used = 0;
}

/*
 * ownership passed into dynbuf
 */
void		dynbuf_add_elt(struct dynbuf	*dbp,
			       void		*elt)
{
	void	*nobj;
	int	nsize;

	if (dbp->used == dbp->size) {
		if (dbp->size == 0)	nsize = DYNBUF_INITIAL_SIZE;
		else			nsize = 2 * dbp->size;
		nobj = realloc(dbp->obj,nsize * sizeof(void *));
		if (!nobj) {
			perror("dynbuf_add_elt");
			desc_printf(2,SHNAME ": out of memory expanding dynbuf\n");
			exit(1);
		}
		dbp->obj = nobj;
		dbp->size = nsize;
	}
	/* used < size */
	dbp->obj[dbp->used++] = elt;
}
int		dynbuf_nelt(struct dynbuf	*dbp)
{
	return dbp->used;
}

void		*dynbuf_get_elt(struct dynbuf	*dbp,
				int		idx)
{
	if ((unsigned) idx < dbp->used)	return dbp->obj[idx];
	else				return 0;
}

void		*dynbuf_take_elt(struct dynbuf	*dbp,
				 int		idx)
{
	void	*rv;

	if ((unsigned) idx < dbp->used)	{
		rv = dbp->obj[idx];
		dbp->obj[idx] = 0;
		return rv;
	}
	return 0;
}

void		*dynbuf_set_elt(struct dynbuf	*dbp,
				int		idx,
				void		*elt)
{
	void	*rv;

	if ((unsigned) idx < dbp->used)	{
		rv = dbp->obj[idx];
		dbp->obj[idx] = elt;
		return rv;
	}
	return 0;
}

int		dynbuf_extend_size(struct dynbuf	*dbp,
				   int			size)
{
	void	*nobj;

	/* extend_size does not shrink! */
	if (size <= dbp->size) return 0;
	nobj = realloc(dbp->obj,size);
	if (!nobj)
		return 0;
	dbp->obj = nobj;
	dbp->size = size;
	return 1;
}

void		**dynbuf_take_buf(struct dynbuf		*dbp)
{
	void	**obj = dbp->obj;

	dbp->size = dbp->used = 0;
	dbp->obj = 0;

	return obj;
}

void		dynbuf_clear(struct dynbuf		*dbp)
{
	dbp->used = 0;
}
